#include "mpi.h"

#define NUM_DIMS 2

worker( argc, argv )
int argc;
char **argv;
{
    int              rank, size, i;
    MPI_Comm_request handle;
    MPI_Status       status;
    int              errors=0;
    int              dims[NUM_DIMS];
    int              periods[NUM_DIMS];
    int              coords[NUM_DIMS];
    int              new_coords[NUM_DIMS];
    int              reorder = 1;
    MPI_Comm comm_cart, new_comm;
    int              topo_status;
    int              ndims;
    int              new_rank;
    int              remain_dims[NUM_DIMS];
    int              newnewrank;

    MPI_Init( &argc, argv );

    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );

    /* Clear dims array and get dims for topology */
    for(i=0;i<NUM_DIMS;i++) { dims[i] = 0; periods[i] = 0; }
    MPI_Make_dims ( size, NUM_DIMS, dims );

    /* Make a new communicator with a topology */
    MPI_Make_cart ( MPI_COMM_WORLD, 2, dims, periods, reorder, &comm_cart );

    /* Determine the status of the new communicator */
    MPI_Topo_status ( comm_cart, &topo_status );
    if (topo_status != MPI_CART) errors++;

    /* How many dims do we have? */
    MPI_Get_cartdim( comm_cart, &ndims );
    if ( ndims != NUM_DIMS ) errors++;

    /* Get the topology, does it agree with what we put in? */
    for(i=0;i<NUM_DIMS;i++) { dims[i] = 0; periods[i] = 0; }
    MPI_Get_cart ( comm_cart, NUM_DIMS, dims, periods, coords );

    /* Does the mapping from coords to rank work? */
    MPI_Cart_rank ( comm_cart, coords, &new_rank );
    if ( new_rank != rank ) errors++;

    /* Does the mapping from rank to coords work */
    MPI_Cart_coords ( comm_cart, rank, NUM_DIMS, new_coords );
    for (i=0;i<NUM_DIMS;i++) 
      if ( coords[i] != new_coords[i] ) 
	errors++;

    /* Let's shift in each dimension and see how it works!   */
    /* Because it's late and I'm tired, I'm not making this  */
    /* automatically test itself.                            */
    for (i=0;i<NUM_DIMS;i++) {
      int source, dest;
      printf ("[%d] Shifting %d in the %d dimension\n",rank,1,i);
      MPI_Cart_shift(comm_cart, i, 1, &source, &dest);
      printf ("[%d]    source = %d  dest = %d\n",rank,source,dest);
    }

    /* Subdivide */
    remain_dims[0] = 0; 
    for (i=1; i<NUM_DIMS; i++) remain_dims[i] = 1;
    MPI_Cart_sub ( comm_cart, remain_dims, &new_comm );

    /* Determine the status of the new communicator */
    MPI_Topo_status ( new_comm, &topo_status );
    if (topo_status != MPI_CART) errors++;

    /* How many dims do we have? */
    MPI_Get_cartdim( new_comm, &ndims );
    if ( ndims != NUM_DIMS-1 ) errors++;

    /* Get the topology, does it agree with what we put in? */
    for(i=0;i<NUM_DIMS-1;i++) { dims[i] = 0; periods[i] = 0; }
    MPI_Get_cart ( new_comm, ndims, dims, periods, coords );
    
    /* Does the mapping from coords to rank work? */
    MPI_Comm_rank ( new_comm, &newnewrank );
    MPI_Cart_rank ( new_comm, coords, &new_rank );
    if ( new_rank != newnewrank ) errors++;

    /* Does the mapping from rank to coords work */
    MPI_Cart_coords ( new_comm, new_rank, NUM_DIMS -1, new_coords );
    for (i=0;i<NUM_DIMS-1;i++) 
      if ( coords[i] != new_coords[i] ) 
	errors++;

    /* We're at the end */
    MPI_Finalize();
    if (errors) printf( "[%d] done with %d ERRORS!\n", rank,errors );
    else        printf( "[%d] SUCCESSFUL!\n", rank );
}
